@@ -18,7 +18,7 @@
         getfunctions(deviceid)
         getproperties(deviceid)
         getdps(deviceid)
-        sendcommand(deviceid, commands)
+        sendcommand(deviceid, commands [, uri])
         getconnectstatus(deviceid)
         getdevicelog(deviceid, start=[now - 1 day], end=[now], evtype="1,2,3,4,5,6,7,8,9,10", size=100, params={})
           -> when start or end are negative, they are the number of days before "right now"
@@ -33,7 +33,12 @@ import hashlib
 import hmac
 import json
 import time
-import requests
+try:
+    import requests
+except ImportError as impErr:
+    print("WARN: Unable to import requests library, Cloud functions will not work.")
+    print("WARN: Check dependencies. See https://github.com/jasonacox/tinytuya/issues/377")
+    print("WARN: Error: {}.".format(impErr.args[0]))
 
 from .core import * # pylint: disable=W0401, W0614
 
@@ -42,7 +47,7 @@ from .core import * # pylint: disable=W0401, W0614
 ########################################################
 
 class Cloud(object):
-    def __init__(self, apiRegion=None, apiKey=None, apiSecret=None, apiDeviceID=None, new_sign_algorithm=True, initial_token=None):
+    def __init__(self, apiRegion=None, apiKey=None, apiSecret=None, apiDeviceID=None, new_sign_algorithm=True, initial_token=None, **extrakw):
         """
         Tuya Cloud IoT Platform Access
 
@@ -83,6 +88,8 @@ class Cloud(object):
         self.error = None
         self.new_sign_algorithm = new_sign_algorithm
         self.server_time_offset = 0
+        self.use_old_device_list = True
+        self.mappings = None
 
         if (not apiKey) or (not apiSecret):
             try:
@@ -93,7 +100,8 @@ class Cloud(object):
                     self.apiRegion = config['apiRegion']
                     self.apiKey = config['apiKey']
                     self.apiSecret = config['apiSecret']
-                    self.apiDeviceID = config['apiDeviceID']
+                    if 'apiDeviceID' in config:
+                        self.apiDeviceID = config['apiDeviceID']
             except:
                 self.error = error_json(
                     ERR_CLOUDKEY,
@@ -125,7 +133,7 @@ class Cloud(object):
         if self.apiRegion == "in":
             self.urlhost = "openapi.tuyain.com"      # India Datacenter
 
-    def _tuyaplatform(self, uri, action='GET', post=None, ver='v1.0', recursive=False, query=None):
+    def _tuyaplatform(self, uri, action='GET', post=None, ver='v1.0', recursive=False, query=None, content_type=None):
         """
         Handle GET and POST requests to Tuya Cloud
         """
@@ -141,9 +149,12 @@ class Cloud(object):
         sign_url = url
         if post is not None:
             body = json.dumps(post)
-            headers['Content-type'] = 'application/json'
         if action not in ('GET', 'POST', 'PUT', 'DELETE'):
             action = 'POST' if post else 'GET'
+        if action == 'POST' and content_type is None:
+            content_type = 'application/json'
+        if content_type:
+            headers['Content-type'] = content_type
         if query:
             # note: signature must be calculated before URL-encoding!
             if type(query) == str:
@@ -196,6 +207,7 @@ class Cloud(object):
         headers['sign'] = signature
         headers['t'] = str(now)
         headers['sign_method'] = 'HMAC-SHA256'
+        headers['mode'] = 'cors'
 
         if self.token is not None:
             headers['access_token'] = self.token
@@ -211,6 +223,9 @@ class Cloud(object):
                 "POST: URL=%s HEADERS=%s DATA=%s", url, headers, body,
             )
             response = requests.post(url, headers=headers, data=body)
+            log.debug(
+                "POST RESPONSE: code=%d text=%s token=%s", response.status_code, response.text, self.token
+            )
 
         # Check to see if token is expired
         if "token invalid" in response.text:
@@ -305,24 +320,141 @@ class Cloud(object):
             action = 'POST' if post else 'GET'
         return self._tuyaplatform(url, action=action, post=post, ver=None, query=query)
 
-    def getdevices(self, verbose=False):
+    # merge device list 'result2' into 'result1'
+    # if result2 has a device which is not in result1 then it will be added
+    # if result2 has a key which does not exist or is empty in result1 then that key will be copied over
+    def _update_device_list( self, result1, result2 ):
+        for new_device in result2:
+            if 'id' not in new_device or not new_device['id']:
+                continue
+            found = False
+            for existing_device in result1:
+                if 'id' in existing_device and existing_device['id'] == new_device['id']:
+                    found = True
+                    for k in new_device:
+                        if k not in existing_device or not existing_device[k]:
+                            existing_device[k] = new_device[k]
+            if not found:
+                result1.append( new_device )
+
+    def _get_all_devices( self, uid=None, device_ids=None ):
+        fetches = 0
+        our_result = { 'result': [] }
+        last_row_key = None
+        has_more = True
+        total = 0
+
+        if uid:
+            # get device list for specified user id
+            query = {'page_size':'75', 'source_type': 'tuyaUser', 'source_id': uid}
+            # API docu: https://developer.tuya.com/en/docs/cloud/dc413408fe?id=Kc09y2ons2i3b
+            uri = '/v1.3/iot-03/devices'
+            if device_ids:
+                if isinstance( device_ids, tuple ) or isinstance( device_ids, list ):
+                    query['device_ids'] = ','.join(device_ids)
+                else:
+                    query['device_ids'] = device_ids
+        else:
+            # get all devices
+            query = {'size':'50'}
+            # API docu: https://developer.tuya.com/en/docs/cloud/fc19523d18?id=Kakr4p8nq5xsc
+            uri = '/v1.0/iot-01/associated-users/devices'
+
+        while has_more:
+            result = self.cloudrequest( uri, query=query )
+            fetches += 1
+            has_more = False
+
+            if type(result) == dict:
+                log.debug( 'Cloud response:' )
+                log.debug( json.dumps( result, indent=2 ) )
+            else:
+                log.debug( 'Cloud response: %r', result )
+
+            # format it the same as before, basically just moves result->devices into result
+            for i in result:
+                if i == 'result':
+                    # by-user-id has the result in 'list' while all-devices has it in 'devices'
+                    if 'list' in result[i] and 'devices' not in result[i]:
+                        our_result[i] += result[i]['list']
+                    elif 'devices' in result[i]:
+                        our_result[i] += result[i]['devices']
+
+                    if 'total' in result[i]: total = result[i]['total']
+                    if 'last_row_key' in result[i]:
+                        query['last_row_key'] = result[i]['last_row_key']
+                    if 'has_more' in result[i]:
+                        has_more = result[i]['has_more']
+                else:
+                    our_result[i] = result[i]
+
+        our_result['fetches'] = fetches
+        our_result['total'] = total
+
+        return our_result
+
+    def getdevices(self, verbose=False, oldlist=[], include_map=False):
         """
         Return dictionary of all devices.
         If verbose is true, return full Tuya device
         details.
         """
-        uid = self._getuid(self.apiDeviceID)
-        if uid is None:
-            return error_json(
-                ERR_CLOUD,
-                "Unable to get uid for device list"
-            )
-        elif isinstance( uid, dict):
-            return uid
+        old_devices = {}
+        if oldlist:
+            for dev in oldlist:
+                dev_id = dev['id']
+                old_devices[dev_id] = dev
 
-        # Use UID to get list of all Devices for User
-        uri = 'users/%s/devices' % uid
-        json_data = self._tuyaplatform(uri)
+        if self.apiDeviceID and self.use_old_device_list:
+            json_data = {}
+            uid_list = {}
+
+            # apiDeviceID can be a comma-separated list, so process them all
+            for dev_id in self.apiDeviceID.split(','):
+                dev_id = dev_id.strip()
+                if not dev_id:
+                    continue
+                uid = self._getuid( dev_id )
+                if not uid:
+                    # no user for this device?
+                    continue
+                if isinstance( uid, dict ):
+                    # it's an error_json dict
+                    return uid
+                else:
+                    uid_list[uid] = True
+
+            if not uid_list:
+                return error_json(
+                    ERR_CLOUD,
+                    "Unable to get uid for device list"
+                )
+
+            for uid in uid_list:
+                # Use UID to get list of all Devices for User
+                uri = 'users/%s/devices' % uid
+                json_run = self._tuyaplatform(uri)
+                # merge the dicts
+                for k in json_run:
+                    if (k not in json_data) or (k != 'result'):
+                        # replace if key is not 'result'
+                        json_data[k] = json_run[k]
+                    else:
+                        # merge 'result' keys
+                        json_data[k] += json_run[k]
+        else:
+            json_data = self._get_all_devices()
+            users = {}
+            # loop through all devices and build a list of user IDs
+            for dev in json_data['result']:
+                if 'uid' in dev:
+                    users[dev['uid']] = True
+            if users:
+                # we have at least 1 user id, so fetch the device list again to make sure we have the local key
+                # this also gets us the gateway_id for child devices
+                for uid in users.keys():
+                    json_data2 = self._get_all_devices( uid=uid )
+                    self._update_device_list( json_data['result'], json_data2['result'] )
 
         if verbose:
             return json_data
@@ -331,30 +463,90 @@ class Cloud(object):
                 ERR_CLOUD,
                 "Unable to get device list"
             )
-        else:
-            # Filter to only Name, ID and Key
-            return self.filter_devices( json_data['result'] )
+
+        self.getdevices_raw = json_data
+        devs = json_data['result']
+        changed_devices = []
+        unchanged_devices = []
+
+        # check to see if anything has changed.  if so, re-download factory-infos and DP mapping
+        for dev in devs:
+            dev_id = dev['id']
+            if dev_id not in old_devices:
+                # newly-added device
+                changed_devices.append( dev )
+                continue
+            old = old_devices[dev_id]
+            if 'key' not in old or old['key'] != dev['local_key']:
+                # local key changed
+                changed_devices.append( dev )
+                continue
+            if (('icon' not in old) and ('icon' in dev)) or (include_map and ('mapping' not in old or old['mapping'] is None)):
+                # icon or mapping added
+                changed_devices.append( dev )
+                continue
+            is_same = True
+            for k in DEVICEFILE_SAVE_VALUES:
+                if k in dev and k != 'icon' and k != 'last_ip' and (k not in old or old[k] != dev[k]):
+                    is_same = False
+                    break
+            if not is_same:
+                changed_devices.append( dev )
+                continue
+            unchanged_devices.append( old )
+
+        if include_map:
+            mappings = self.getmappings( changed_devices )
+            for productid in mappings:
+                for dev in changed_devices:
+                    if 'product_id' in dev and dev['product_id'] == productid:
+                        dev['mapping'] = mappings[productid]
+                # also set unchanged devices just in case the mapping changed
+                for dev in unchanged_devices:
+                    if 'product_id' in dev and dev['product_id'] == productid:
+                        dev['mapping'] = mappings[productid]
+
+        log.debug( 'changed: %d', len(changed_devices) )
+        log.debug( 'unchanged: %d', len(unchanged_devices) )
+
+        # Filter to only Name, ID and Key
+        return self.filter_devices( changed_devices ) + unchanged_devices
+
+    def _get_hw_addresses( self, maclist, devices ):
+        while devices:
+            # returns id, mac, uuid (and sn if available)
+            uri = 'devices/factory-infos?device_ids=%s' % (",".join(devices[:50]))
+            result = self._tuyaplatform(uri)
+            log.debug( json.dumps( result, indent=2 ) )
+            if 'result' in result:
+                for dev in result['result']:
+                    if 'id' in dev:
+                        dev_id = dev['id']
+                        del dev['id']
+                        maclist[dev_id] = dev
+            devices = devices[50:]
 
     def filter_devices( self, devs, ip_list=None ):
-        # Use Device ID to get MAC addresses
-        uri = 'devices/factory-infos?device_ids=%s' % (",".join(i['id'] for i in devs))
-        json_mac_data = self._tuyaplatform(uri)
+        json_mac_data = {}
+        # mutable json_mac_data will be modified
+        self._get_hw_addresses( json_mac_data, [i['id'] for i in devs] )
 
         tuyadevices = []
         icon_host = 'https://images.' + self.urlhost.split( '.', 1 )[1] + '/'
 
         for i in devs:
-            item = {}
-            item['name'] = i['name'].strip()
-            item['id'] = i['id']
-            item['key'] = i['local_key']
-            if 'mac' in i:
-                item['mac'] = i['mac']
-            else:
-                try:
-                    item['mac'] = next((m['mac'] for m in json_mac_data['result'] if m['id'] == i['id']), "")
-                except:
-                    pass
+            dev_id = i['id']
+            item = {
+                'name': '' if 'name' not in i else i['name'].strip(),
+                'id': dev_id,
+                'key': '' if 'local_key' not in i else i['local_key'],
+                'mac': '' if 'mac' not in i else i['mac']
+            }
+
+            if dev_id in json_mac_data:
+                for k in ('mac','uuid','sn'):
+                    if k in json_mac_data[dev_id]:
+                        item[k] = json_mac_data[dev_id][k]
 
             if ip_list and 'mac' in item and item['mac'] in ip_list:
                 item['ip'] = ip_list[item['mac']]
@@ -366,6 +558,10 @@ class Cloud(object):
                     else:
                         item[k] = i[k]
 
+            if 'gateway_id' in i:
+                k = 'gateway_id'
+                item[k] = i[k]
+
             tuyadevices.append(item)
 
         return tuyadevices
@@ -425,7 +621,7 @@ class Cloud(object):
             )
         return response_dict
 
-    def sendcommand(self, deviceid=None, commands=None):
+    def sendcommand(self, deviceid=None, commands=None, uri='iot-03/devices/'):
         """
         Send a command to the device
         """
@@ -436,7 +632,7 @@ class Cloud(object):
                 ERR_PARAMS,
                 "Missing DeviceID and/or Command Parameters"
             )
-        uri = 'iot-03/devices/%s/commands' % (deviceid)
+        uri += '%s/commands' % (deviceid)
         response_dict = self._tuyaplatform(uri,action='POST',post=commands)
 
         if not response_dict['success']:
@@ -598,3 +794,94 @@ class Cloud(object):
         elif len(str(ts)) == 10:
             ts *= 1000
         return ts
+
+    @staticmethod
+    def _build_mapping( src, dst ):
+        # merge multiple DPS sets from result['status'] and result['functions'] into a single result
+        for mapp in src:
+            try:
+                code = mapp['code']
+                dp_id = code if 'dp_id' not in mapp else str(mapp['dp_id'])
+                if dp_id in dst:
+                    continue
+                data = { 'code': code, 'type': mapp['type'] }
+                if mapp['type'].lower() == 'string':
+                    values = mapp['values']
+                else:
+                    try:
+                        values = json.loads( mapp['values'] )
+                    except:
+                        values = mapp['values']
+                if values and type(values) == dict and 'unit' in values:
+                    if values['unit']:
+                        # normalize some unicode and temperature units
+                        # not sure what a good replacement for '份' is (seen in a pet feeder)
+                        values['unit'] = values['unit'].replace('℉','°F').replace('℃','°C').replace('f','°F').replace('c','°C').replace('秒','s')
+
+                # Tuya's 'JSON' mapping type is an ordered dict, but python's dict is not!  so, save the raw value as well
+                if mapp['type'].lower() == 'json':
+                    data['raw_values'] = mapp['values']
+                data['values'] = values
+                dst[dp_id] = data
+            except:
+                log.debug( 'Parse mapping item failed!', exc_info=True )
+
+    def getmapping( self, productid, deviceid=None ):
+        # return a mapping for the given product id, or download it from the cloud using a device id
+        # Return value: None on failure, or a dict on success (may be an empty dict if device does not have DPs)
+        if not self.mappings:
+            self.mappings = {} #load_mappings()
+
+        if productid in self.mappings:
+            # already have this product id, so just return it
+            return self.mappings[productid]
+
+        if deviceid:
+            # we do not have this product id yet, so download it via this device id
+            result = self.getdps(deviceid)
+
+            if result:
+                if 'result' in result:
+                    result = result['result']
+                    dps = {}
+                    # merge result['status'] and result['functions'] into a single result
+                    if 'status' in result:
+                        self._build_mapping( result['status'], dps )
+                    if 'functions' in result:
+                        self._build_mapping( result['functions'], dps )
+                    self.mappings[productid] = dps
+                    log.debug( 'Downloaded mapping for device %r: %r', deviceid, dps)
+                elif ('code' in result and result['code'] == 2009) or ('msg' in result and result['msg'] == 'not support this device'):
+                    # this device does not have any DPs!
+                    self.mappings[productid] = {}
+
+        if productid in self.mappings:
+            # download success, so return it
+            return self.mappings[productid]
+
+        # no device id, or download failed
+        return None
+
+    def setmappings( self, mappings ):
+        # sets an initial mapping set so we do not need to download everything
+        if type(mappings) == dict:
+            self.mappings = mappings
+
+    def getmappings( self, devices ):
+        # get all mappings for all tuya devices
+        # returns a dict with product ids as keys
+        if not self.mappings:
+            self.mappings = {}
+
+        for dev in devices:
+            try:
+                devid = dev['id']
+                productid = dev['product_id']
+            except:
+                # we need both the device id and the product id to download mappings!
+                continue
+
+            if productid not in self.mappings:
+                self.getmapping( productid, devid )
+
+        return self.mappings
