@@ -115,12 +115,11 @@ import json
 import logging
 import struct
 import time
-from math import ceil, floor
 
 from ..core import Device, log, CONTROL
 
 class IRRemoteControlDevice(Device):
-    CMD_SEND_KEY_CODE =	"send_ir"  # Command to start sending a key
+    CMD_SEND_KEY_CODE = "send_ir"   # Command to start sending a key
     DP_SEND_IR        = "201"       # ir_send, send and report (read-write)
     DP_LEARNED_ID     = "202"       # ir_study_code, report only (read-only)
     DP_MODE           =   "1"
@@ -242,7 +241,7 @@ class IRRemoteControlDevice(Device):
     def study_start( self ):
         self.send_command( 'study' )
 
-    def	study_end( self ):
+    def study_end( self ):
         self.send_command( 'study_exit' )
 
     def receive_button( self, timeout=30 ):
@@ -313,7 +312,7 @@ class IRRemoteControlDevice(Device):
     @staticmethod
     def build_head( freq=38, bit_time=0, zero_time=0, one_time=0, bit_time_type=1, timings=[], convert_time=True ):
         timings = list(timings)
-        freq =	round( freq * 100)
+        freq = round( freq * 100)
         if not bit_time and len(timings) > 0:
             bit_time = timings[0]
             timings = timings[1:]
@@ -461,7 +460,7 @@ class IRRemoteControlDevice(Device):
                     data += '0'
             else:
                 # if the first byte is 0, the next byte is how many bits to transmit
-                byts = ceil( bits / 8 )
+                byts = int( (bits + 7) / 8 )
                 cnt = byts * 2
                 data = key[:cnt]
 
@@ -794,7 +793,7 @@ class IRRemoteControlDevice(Device):
                             bits = data = 0
                     elif k == one_symbol:
                         removed += k_symbol_pattern
-                        bits +=	1
+                        bits += 1
                         data |= 1 << (8 - bits)
                         if bits == 8:
                             byts.append( data )
@@ -869,7 +868,7 @@ class IRRemoteControlDevice(Device):
         return header, '01' + key1
 
     @staticmethod
-    def	_merge_similar_pulse_times( p_count, fudge ):
+    def _merge_similar_pulse_times( p_count, fudge ):
         p_map = { }
         mod = True
         while mod:
@@ -1028,7 +1027,7 @@ class IRRemoteControlDevice(Device):
         result_string = ''
         symbols = { results[0][0]: '%', 4500: '^', 2250: '&', results[0][5]: '*' }
         for r in results:
-            if r[0] not in symbols or r[1] not	in symbols:
+            if r[0] not in symbols or r[1] not in symbols:
                 return None
             result_string += symbols[r[0]] + symbols[r[1]]
             result_string += IRRemoteControlDevice._build_key_bitfield( r[3], r[4], r[2] )
