@@ -10,11 +10,26 @@
 
  Classes
   * AESCipher - Cryptography Helpers
-  * XenonDevice(dev_id, address=None, local_key="", dev_type="default", connection_timeout=5, version="3.1", persist=False) - Base Tuya Objects and Functions
+  * XenonDevice(...) - Base Tuya Objects and Functions
+        XenonDevice(dev_id, address=None, local_key="", dev_type="default", connection_timeout=5, 
+            version="3.1", persist=False, cid/node_id=None, parent=None, connection_retry_limit=5, 
+            connection_retry_delay=5)
   * Device(XenonDevice) - Tuya Class for Devices
 
- Functions
+ Module Functions
+    set_debug(toggle, color)                    # Activate verbose debugging output
+    pack_message(msg, hmac_key=None)            # Packs a TuyaMessage() into a network packet, encrypting or adding a CRC if protocol requires
+    unpack_message(data, hmac_key=None, header=None, no_retcode=False)
+                                                # Unpacks a TuyaMessage() from a network packet, decrypting or checking the CRC if protocol requires
+    parse_header(data)                          # Unpacks just the header part of a message into a TuyaHeader()
+    find_device(dev_id=None, address=None)      # Scans network for Tuya devices with either ID = dev_id or IP = address
+    device_info(dev_id)                         # Searches DEVICEFILE (usually devices.json) for devices with ID = dev_id and returns just that device
+    assign_dp_mappings(tuyadevices, mappings)   # Adds mappings to all the devices in the tuyadevices list
+    decrypt_udp(msg)                            # Decrypts a UDP network broadcast packet
+
+ Device Functions
     json = status()                    # returns json payload
+    subdev_query(nowait)               # query sub-device status (only for gateway devices)
     set_version(version)               # 3.1 [default], 3.2, 3.3 or 3.4
     set_socketPersistent(False/True)   # False [default] or True
     set_socketNODELAY(False/True)      # False or True [default]
@@ -33,10 +48,9 @@
     turn_on(switch=1, nowait)          # Turn on device / switch #
     turn_off(switch=1, nowait)         # Turn off
     set_timer(num_secs, nowait)        # Set timer for num_secs
-    set_debug(toggle, color)           # Activate verbose debugging output
     set_sendWait(num_secs)             # Time to wait after sending commands before pulling response
     detect_available_dps()             # Return list of DPS available from device
-    generate_payload(command, data)    # Generate TuyaMessage payload for command with data
+    generate_payload(command, data,...)# Generate TuyaMessage payload for command with data
     send(payload)                      # Send payload to device (do not wait for response)
     receive()                          # Receive payload from device
 
@@ -47,8 +61,8 @@
     The origin of this python module (now abandoned)
   * LocalTuya https://github.com/rospogrigio/localtuya-homeassistant by rospogrigio
     Updated pytuya to support devices with Device IDs of 22 characters
-  * Tuya Protocol 3.4 Support by uzlonewolf
-    Enhancement to TuyaMessage logic for multi-payload messages and Tuya Protocol 3.4 support
+  * Tuya Protocol 3.4 and 3.5 Support by uzlonewolf
+    Enhancement to TuyaMessage logic for multi-payload messages
 
 """
 
@@ -85,22 +99,11 @@ except ImportError:
 # Colorama terminal color capability for all platforms
 init()
 
-version_tuple = (1, 10, 3)
+version_tuple = (1, 12, 11)
 version = __version__ = "%d.%d.%d" % version_tuple
 __author__ = "jasonacox"
 
 log = logging.getLogger(__name__)
-# Uncomment the following to set debug mode or call set_debug()
-# logging.basicConfig(level=logging.DEBUG)
-
-log.debug("%s version %s", __name__, __version__)
-log.debug("Python %s on %s", sys.version, sys.platform)
-if Crypto is None:
-    log.debug("Using pyaes version %r", pyaes.VERSION)  # pylint: disable=E0601
-    log.debug("Using pyaes from %r", pyaes.__file__)    # pylint: disable=E0601
-else:
-    log.debug("Using PyCrypto %r", Crypto.version_info)
-    log.debug("Using PyCrypto from %r", Crypto.__file__)
 
 # Globals Network Settings
 MAXCOUNT = 15       # How many tries before stopping
@@ -119,7 +122,7 @@ DEVICEFILE = 'devices.json'
 RAWFILE = 'tuya-raw.json'
 SNAPSHOTFILE = 'snapshot.json'
 
-DEVICEFILE_SAVE_VALUES = ('category', 'product_name', 'product_id', 'biz_type', 'model', 'sub', 'icon', 'version', 'last_ip', 'uuid', 'node_id')
+DEVICEFILE_SAVE_VALUES = ('category', 'product_name', 'product_id', 'biz_type', 'model', 'sub', 'icon', 'version', 'last_ip', 'uuid', 'node_id', 'sn', 'mapping')
 
 # Tuya Command Types
 # Reference: https://github.com/tuya/tuya-iotos-embeded-sdk-wifi-ble-bk7231n/blob/master/sdk/include/lan_protocol.h
@@ -170,7 +173,7 @@ PREFIX_6699_BIN = b"\x00\x00\x66\x99"
 SUFFIX_6699_VALUE = 0x00009966
 SUFFIX_6699_BIN = b"\x00\x00\x99\x66"
 
-NO_PROTOCOL_HEADER_CMDS = [DP_QUERY, DP_QUERY_NEW, UPDATEDPS, HEART_BEAT, SESS_KEY_NEG_START, SESS_KEY_NEG_RESP, SESS_KEY_NEG_FINISH ]
+NO_PROTOCOL_HEADER_CMDS = [DP_QUERY, DP_QUERY_NEW, UPDATEDPS, HEART_BEAT, SESS_KEY_NEG_START, SESS_KEY_NEG_RESP, SESS_KEY_NEG_FINISH, LAN_EXT_STREAM ]
 
 # Python 2 Support
 IS_PY2 = sys.version_info[0] == 2
@@ -198,6 +201,7 @@ ERR_CLOUDRESP = 910
 ERR_CLOUDTOKEN = 911
 ERR_PARAMS = 912
 ERR_CLOUD = 913
+ERR_KEY_OR_VER = 914
 
 error_codes = {
     ERR_JSON: "Invalid JSON Response from Device",
@@ -214,6 +218,7 @@ error_codes = {
     ERR_CLOUDTOKEN: "Unable to Get Cloud Token",
     ERR_PARAMS: "Missing Function Parameters",
     ERR_CLOUD: "Error Response from Tuya Cloud",
+    ERR_KEY_OR_VER: "Check device key or version",
     None: "Unknown Error",
 }
 
@@ -249,6 +254,7 @@ class AESCipher(object):
                 raise NotImplementedError( 'pyaes does not support GCM, please install PyCryptodome' )
 
             _ = self._pad(raw)
+            # pylint: disable-next=used-before-assignment
             cipher = pyaes.blockfeeder.Encrypter(
                 pyaes.AESModeOfOperationECB(self.key),
                 pyaes.PADDING_DEFAULT if pad else pyaes.PADDING_NONE
@@ -341,6 +347,12 @@ def set_debug(toggle=True, color=True):
             logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.DEBUG)
         log.setLevel(logging.DEBUG)
         log.debug("TinyTuya [%s]\n", __version__)
+        log.debug("Python %s on %s", sys.version, sys.platform)
+        if Crypto is None:
+            # pylint: disable-next=used-before-assignment
+            log.debug("Using pyaes version %r", pyaes.VERSION)
+        else:
+            log.debug("Using PyCrypto %r", Crypto.version_info)
     else:
         log.setLevel(logging.NOTSET)
 
@@ -520,14 +532,26 @@ def find_device(dev_id=None, address=None):
     if dev_id is None and address is None:
         return (None, None, None)
     log.debug("Listening for device %s on the network", dev_id)
+
     # Enable UDP listening broadcasting mode on UDP port 6666 - 3.1 Devices
     client = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
     client.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
+    try:
+        client.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
+    except AttributeError:
+        # SO_REUSEPORT not available
+        pass
     client.bind(("", UDPPORT))
     client.setblocking(False)
+
     # Enable UDP listening broadcasting mode on encrypted UDP port 6667 - 3.3 Devices
     clients = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
     clients.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
+    try:
+        clients.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
+    except AttributeError:
+        # SO_REUSEPORT not available
+        pass
     clients.bind(("", UDPPORTS))
     clients.setblocking(False)
 
@@ -598,7 +622,7 @@ def device_info( dev_id ):
         with open(DEVICEFILE, 'r') as f:
             tuyadevices = json.load(f)
             log.debug("loaded=%s [%d devices]", DEVICEFILE, len(tuyadevices))
-            for	dev in tuyadevices:
+            for dev in tuyadevices:
                 if 'id' in dev and dev['id'] == dev_id:
                     log.debug("Device %r found in %s", dev_id, DEVICEFILE)
                     devinfo = dev
@@ -609,6 +633,37 @@ def device_info( dev_id ):
 
     return devinfo
 
+def assign_dp_mappings( tuyadevices, mappings ):
+    """ Adds mappings to all the devices in the tuyadevices list
+
+    Parameters:
+        tuyadevices = list of devices
+        mappings = dict containing the mappings
+
+    Response:
+        Nothing, modifies tuyadevices in place
+    """
+    if type(mappings) != dict:
+        raise ValueError( '\'mappings\' must be a dict' )
+
+    if (not mappings) or (not tuyadevices):
+        return None
+
+    for dev in tuyadevices:
+        try:
+            devid = dev['id']
+            productid = dev['product_id']
+        except:
+            # we need both the device id and the product id to download mappings!
+            log.debug( 'Cannot add DP mapping, no device id and/or product id: %r', dev )
+            continue
+
+        if productid in mappings:
+            dev['mapping'] = mappings[productid]
+        else:
+            log.debug( 'Device %s has no mapping!', devid )
+            dev['mapping'] = None
+
 # Tuya Device Dictionary - Command and Payload Overrides
 #
 # 'default' devices require the 0a command for the DP_QUERY request
@@ -637,6 +692,7 @@ payload_dict = {
         CONTROL_NEW: {"command": {"devId": "", "uid": "", "t": ""}},
         DP_QUERY_NEW: {"command": {"devId": "", "uid": "", "t": ""}},
         UPDATEDPS: {"command": {"dpId": [18, 19, 20]}},
+        LAN_EXT_STREAM: { "command": { "reqType": "", "data": {} }},
     },
     # Special Case Device with 22 character ID - Some of these devices
     # Require the 0d command as the DP_QUERY status request and the list of
@@ -647,62 +703,107 @@ payload_dict = {
             "command": {"devId": "", "uid": "", "t": ""},
         },
     },
+    # v3.3+ devices do not need devId/gwId/uid
     "v3.4": {
         CONTROL: {
             "command_override": CONTROL_NEW,  # Uses CONTROL_NEW command
-            "command": {"protocol":5, "t": "int", "data": ""}
+            "command": {"protocol":5, "t": "int", "data": {}}
             },
-        DP_QUERY: { "command_override": DP_QUERY_NEW },
+        CONTROL_NEW: {
+            "command": {"protocol":5, "t": "int", "data": {}}
+        },
+        DP_QUERY: {
+            "command_override": DP_QUERY_NEW,
+            "command": {} #"protocol":4, "t": "int", "data": {}}
+        },
+        DP_QUERY_NEW: {
+            "command": {}
+        },
     },
+    # v3.5 is just a copy of v3.4
     "v3.5": {
         CONTROL: {
             "command_override": CONTROL_NEW,  # Uses CONTROL_NEW command
-            "command": {"protocol":5, "t": "int", "data": ""}
+            "command": {"protocol":5, "t": "int", "data": {}}
+        },
+        CONTROL_NEW: {
+            "command": {"protocol":5, "t": "int", "data": {}}
+        },
+        DP_QUERY: {
+            "command_override": DP_QUERY_NEW,
+            "command": {}
+        },
+        DP_QUERY_NEW: {
+            "command": {}
         },
-        DP_QUERY: { "command_override": DP_QUERY_NEW },
     },
+    # placeholders, not yet needed
+    "gateway": { },
+    "gateway_v3.4": { },
+    "gateway_v3.5": { },
     "zigbee": {
+        CONTROL: { "command": {"t": "int", "cid": ""} },
+        DP_QUERY: { "command": {"t": "int", "cid": ""} },
+    },
+    "zigbee_v3.4": {
         CONTROL: {
-            "command": {"t": ""},
+            "command_override": CONTROL_NEW,
+            "command": {"protocol":5, "t": "int", "data": {"cid":""}}
         },
-    }
+        CONTROL_NEW: {
+            "command": {"protocol":5, "t": "int", "data": {"cid":""}}
+        },
+    },
+    "zigbee_v3.5": {
+        CONTROL: {
+            "command_override": CONTROL_NEW,
+            "command": {"protocol":5, "t": "int", "data": {"cid":""}}
+        },
+        CONTROL_NEW: {
+            "command": {"protocol":5, "t": "int", "data": {"cid":""}}
+        },
+    },
 }
 
-
 ########################################################
 #             Core Classes and Functions
 ########################################################
 
 class XenonDevice(object):
     def __init__(
-            self, dev_id, address=None, local_key="", dev_type="default", connection_timeout=5, version=3.1, persist=False, cid=None, parent=None # pylint: disable=W0621
+            self, dev_id, address=None, local_key="", dev_type="default", connection_timeout=5, version=3.1, persist=False, cid=None, node_id=None, parent=None, connection_retry_limit=5, connection_retry_delay=5 # pylint: disable=W0621
     ):
         """
         Represents a Tuya device.
 
         Args:
             dev_id (str): The device id.
-            cid (str: Optional sub device id. Default to None.
             address (str): The network address.
             local_key (str, optional): The encryption key. Defaults to None.
+            cid (str: Optional sub device id. Default to None.
+            node_id (str: alias for cid)
+            parent (object: gateway device this device is a child of)
 
         Attributes:
             port (int): The port to connect to.
         """
 
         self.id = dev_id
-        self.cid = cid
+        self.cid = cid if cid else node_id
         self.address = address
+        self.auto_ip = False
+        self.dev_type = dev_type
+        self.dev_type_auto = self.dev_type == 'default'
+        self.last_dev_type = ''
         self.connection_timeout = connection_timeout
         self.retry = True
-        self.dev_type = dev_type
         self.disabledetect = False  # if True do not detect device22
         self.port = TCPPORT  # default - do not expect caller to pass in
         self.socket = None
         self.socketPersistent = False if not persist else True # pylint: disable=R1719
         self.socketNODELAY = True
-        self.socketRetryLimit = 5
-        self.socketRetryDelay = 5
+        self.socketRetryLimit = connection_retry_limit
+        self.socketRetryDelay = connection_retry_delay
         self.version = 0
         self.dps_to_request = {}
         self.seqno = 1
@@ -713,21 +814,34 @@ class XenonDevice(object):
         self.received_wrong_cid_queue = []
         self.local_nonce = b'0123456789abcdef' # not-so-random random key
         self.remote_nonce = b''
+        self.payload_dict = None
 
         if not local_key:
             local_key = ""
-            devinfo = device_info( dev_id )
-            if devinfo and 'key' in devinfo and devinfo['key']:
-                local_key = devinfo['key']
+            # sub-devices do not need a local key, so only look it up if we are not a sub-device
+            if not parent:
+                devinfo = device_info( dev_id )
+                if devinfo and 'key' in devinfo and devinfo['key']:
+                    local_key = devinfo['key']
         self.local_key = local_key.encode("latin1")
         self.real_local_key = self.local_key
         self.cipher = None
 
         if self.parent:
+            # if we are a child then we should have a cid/node_id but none were given - try and find it the same way we look up local keys
+            if not self.cid:
+                devinfo = device_info( dev_id )
+                if devinfo and 'node_id' in devinfo and devinfo['node_id']:
+                    self.cid = devinfo['node_id']
+            if not self.cid:
+                # not fatal as the user could have set the device_id to the cid
+                # in that case dev_type should be 'zigbee' to set the proper fields in requests
+                log.debug( 'Child device but no cid/node_id given!' )
             XenonDevice.set_version(self, self.parent.version)
             self.parent._register_child(self)
         elif (not address) or address == "Auto" or address == "0.0.0.0":
             # try to determine IP address automatically
+            self.auto_ip = True
             bcast_data = find_device(dev_id)
             if bcast_data['ip'] is None:
                 log.debug("Unable to find device on network (specify IP address)")
@@ -742,9 +856,6 @@ class XenonDevice(object):
             # them (such as BulbDevice) make connections when called
             XenonDevice.set_version(self, 3.1)
 
-        if cid and self.dev_type == 'default':
-            self.dev_type = 'zigbee'
-
     def __del__(self):
         # In case we have a lingering socket connection, close it
         try:
@@ -772,7 +883,28 @@ class XenonDevice(object):
         if self.socket is None:
             # Set up Socket
             retries = 0
+            err = ERR_OFFLINE
             while retries < self.socketRetryLimit:
+                if self.auto_ip and not self.address:
+                    bcast_data = find_device(self.id)
+                    if bcast_data['ip'] is None:
+                        log.debug("Unable to find device on network (specify IP address)")
+                        return ERR_OFFLINE
+                    self.address = bcast_data['ip']
+                    new_version = float(bcast_data['version'])
+                    if new_version != self.version:
+                        # this may trigger a network call which will call _get_socket() again
+                        #self.set_version(new_version)
+                        self.version = new_version
+                        self.version_str = "v" + str(version)
+                        self.version_bytes = str(version).encode('latin1')
+                        self.version_header = self.version_bytes + PROTOCOL_3x_HEADER
+                        self.payload_dict = None
+
+                if not self.address:
+                    log.debug("No address for device!")
+                    return ERR_OFFLINE
+
                 self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                 if self.socketNODELAY:
                     self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
@@ -784,27 +916,36 @@ class XenonDevice(object):
                         # restart session key negotiation
                         if self._negotiate_session_key():
                             return True
+                        else:
+                            if self.socket:
+                                self.socket.close()
+                                self.socket = None
+                            return ERR_KEY_OR_VER
                     else:
                         return True
-                except socket.timeout as err:
+                except socket.timeout as e:
                     # unable to open socket
                     log.debug(
                         "socket unable to connect (timeout) - retry %d/%d",
                         retries, self.socketRetryLimit
                     )
-                except Exception as err:
+                    err = ERR_OFFLINE
+                except Exception as e:
                     # unable to open socket
                     log.debug(
                         "socket unable to connect (exception) - retry %d/%d",
                         retries, self.socketRetryLimit, exc_info=True
                     )
+                    err = ERR_CONNECT
                 if self.socket:
                     self.socket.close()
                     self.socket = None
                 if retries < self.socketRetryLimit:
                     time.sleep(self.socketRetryDelay)
+                if self.auto_ip:
+                    self.address = None
             # unable to get connection
-            return False
+            return err
         # existing socket active
         return True
 
@@ -825,7 +966,8 @@ class XenonDevice(object):
                 tries -= 1
                 if tries == 0:
                     raise DecodeError('No data received - connection closed?')
-                time.sleep(0.1)
+                if self.sendWait is not None:
+                    time.sleep(self.sendWait)
                 continue
             data += newdata
             length -= len(newdata)
@@ -879,7 +1021,7 @@ class XenonDevice(object):
             return self.parent._send_receive_quick(payload, recv_retries, from_child=self)
 
         log.debug("sending payload quick")
-        if not self._get_socket(False):
+        if self._get_socket(False) is not True:
             return None
         enc_payload = self._encode_message(payload) if type(payload) == MessagePayload else payload
         try:
@@ -940,17 +1082,19 @@ class XenonDevice(object):
         msg = None
         while not success:
             # open up socket if device is available
-            if not self._get_socket(False):
+            sock_result = self._get_socket(False)
+            if sock_result is not True:
                 # unable to get a socket - device likely offline
                 self._check_socket_close(True)
-                return error_json(ERR_OFFLINE)
+                return error_json( sock_result if sock_result else ERR_OFFLINE )
             # send request to device
             try:
                 if payload is not None and do_send:
                     log.debug("sending payload")
                     enc_payload = self._encode_message(payload) if type(payload) == MessagePayload else payload
                     self.socket.sendall(enc_payload)
-                    time.sleep(self.sendWait)  # give device time to respond
+                    if self.sendWait is not None:
+                        time.sleep(self.sendWait)  # give device time to respond
                 if getresponse:
                     do_send = False
                     rmsg = self._receive()
@@ -968,7 +1112,7 @@ class XenonDevice(object):
                     else:
                         success = True
                         log.debug("received message=%r", msg)
-                if not getresponse:
+                else:
                     # legacy/default mode avoids persisting socket across commands
                     self._check_socket_close()
                     return None
@@ -983,6 +1127,7 @@ class XenonDevice(object):
                     return None
                 do_send = True
                 retries += 1
+                # toss old socket and get new one
                 self._check_socket_close(True)
                 log.debug(
                     "Timeout in _send_receive() - retry %s / %s",
@@ -995,13 +1140,9 @@ class XenonDevice(object):
                         self.socketRetryLimit
                     )
                     # timeout reached - return error
-                    json_payload = error_json(
-                        ERR_TIMEOUT, "Check device key or version"
-                    )
-                    return json_payload
-                # retry:  wait a bit, toss old socket and get new one
+                    return error_json(ERR_KEY_OR_VER)
+                # wait a bit before retrying
                 time.sleep(0.1)
-                self._get_socket(True)
             except DecodeError as err:
                 log.debug("Error decoding received data - read retry %s/%s", recv_retries, max_recv_retries, exc_info=True)
                 recv_retries += 1
@@ -1017,6 +1158,7 @@ class XenonDevice(object):
                 # likely network or connection error
                 do_send = True
                 retries += 1
+                # toss old socket and get new one
                 self._check_socket_close(True)
                 log.debug(
                     "Network connection error in _send_receive() - retry %s/%s",
@@ -1030,11 +1172,9 @@ class XenonDevice(object):
                     )
                     log.debug("Unable to connect to device ")
                     # timeout reached - return error
-                    json_payload = error_json(ERR_CONNECT)
-                    return json_payload
-                # retry:  wait a bit, toss old socket and get new one
+                    return error_json(ERR_CONNECT)
+                # wait a bit before retrying
                 time.sleep(0.1)
-                self._get_socket(True)
             # except
         # while
 
@@ -1081,18 +1221,21 @@ class XenonDevice(object):
 
         found_child = False
         if self.children:
-            found_cid = want_cid = None
+            found_cid = None
             if result and 'cid' in result:
                 found_cid = result['cid']
+            elif result and 'data' in result and type(result['data']) == dict and 'cid' in result['data']:
+                found_cid = result['data']['cid']
+
+            if found_cid:
                 for c in self.children:
-                    if self.children[c].cid == result['cid']:
-                        want_cid = self.children[c].cid
+                    if self.children[c].cid == found_cid:
                         result['device'] = found_child = self.children[c]
                         break
 
             if from_child and from_child is not True and from_child != found_child:
                 # async update from different CID, try again
-                log.debug( 'Recieved async update for wrong CID %s while looking for CID %s, trying again', found_cid, want_cid )
+                log.debug( 'Recieved async update for wrong CID %s while looking for CID %s, trying again', found_cid, from_child.cid )
                 if self.socketPersistent:
                     # if persistent, save response until the next receive() call
                     # otherwise, trash it
@@ -1101,8 +1244,8 @@ class XenonDevice(object):
                     else:
                         result = self._process_response(result)
                     self.received_wrong_cid_queue.append( (found_child, result) )
-                # do not pass from_child this time to make sure we do not get stuck in a loop
-                return self._send_receive( None, minresponse, True, decode_response, from_child=True)
+                # events should not be coming in so fast that we will never timeout a read, so don't worry about loops
+                return self._send_receive( None, minresponse, True, decode_response, from_child=from_child)
 
         # legacy/default mode avoids persisting socket across commands
         self._check_socket_close()
@@ -1211,7 +1354,7 @@ class XenonDevice(object):
 
         return MessagePayload(SESS_KEY_NEG_START, self.local_nonce)
 
-    def	_negotiate_session_key_generate_step_3( self, rkey ):
+    def _negotiate_session_key_generate_step_3( self, rkey ):
         if not rkey or type(rkey) != TuyaMessage or len(rkey.payload) < 48:
             # error
             log.debug("session key negotiation failed on step 1")
@@ -1334,8 +1477,9 @@ class XenonDevice(object):
         if child.id in self.children and child != self.children[child.id]:
             log.debug('Replacing existing child %r!', child.id)
         self.children[child.id] = child
-        # disable device22 detection as gateways return "json obj data unvalid" when the gateway is polled without a cid
+        # disable device22 detection as some gateways return "json obj data unvalid" when the gateway is polled without a cid
         self.disabledetect = True
+        self.payload_dict = None
 
     def receive(self):
         """
@@ -1367,9 +1511,17 @@ class XenonDevice(object):
                 log.debug("status() rebuilding payload for device22")
                 payload = self.generate_payload(query_type)
                 data = self._send_receive(payload)
+            elif data["Err"] == str(ERR_PAYLOAD):
+                log.debug("Status request returned an error, is version %r and local key %r correct?", self.version, self.local_key)
 
         return data
 
+    def subdev_query( self, nowait=False ):
+        """Query for a list of sub-devices and their status"""
+        # final payload should look like: {"data":{"cids":[]},"reqType":"subdev_online_stat_query"}
+        payload = self.generate_payload(LAN_EXT_STREAM, rawData={"cids":[]}, reqType='subdev_online_stat_query')
+        return self._send_receive(payload, 0, getresponse=(not nowait))
+
     def detect_available_dps(self):
         """Return which datapoints are supported by the device."""
         # device22 devices need a sort of bruteforce querying in order to detect the
@@ -1389,8 +1541,9 @@ class XenonDevice(object):
             except Exception as ex:
                 log.exception("Failed to get status: %s", ex)
                 raise
-            if "dps" in data:
-                self.dps_cache.update(data["dps"])
+            if data is not None and "dps" in data:
+                for k in data["dps"]:
+                    self.dps_cache[k] = None
 
             if self.dev_type == "default":
                 self.dps_to_request = self.dps_cache
@@ -1407,18 +1560,15 @@ class XenonDevice(object):
             self.dps_to_request.update({str(index): None for index in dp_indicies})
 
     def set_version(self, version): # pylint: disable=W0621
-        last_version = self.version
         self.version = version
+        self.version_str = "v" + str(version)
         self.version_bytes = str(version).encode('latin1')
         self.version_header = self.version_bytes + PROTOCOL_3x_HEADER
+        self.payload_dict = None
         if version == 3.2: # 3.2 behaves like 3.3 with device22
             self.dev_type="device22"
             if self.dps_to_request == {}:
                 self.detect_available_dps()
-        elif version >= 3.4:
-            self.dev_type = "v" + str(version)
-        elif last_version >= 3.4 and self.dev_type[0] == "v":
-            self.dev_type = "default"
 
     def set_socketPersistent(self, persist):
         self.socketPersistent = persist
@@ -1472,7 +1622,7 @@ class XenonDevice(object):
         bcast_data = find_device(dev_id=did)
         return (bcast_data['ip'], bcast_data['version'])
 
-    def generate_payload(self, command, data=None, gwId=None, devId=None, uid=None):
+    def generate_payload(self, command, data=None, gwId=None, devId=None, uid=None, rawData=None, reqType=None):
         """
         Generate the payload to send.
 
@@ -1485,19 +1635,63 @@ class XenonDevice(object):
             devId(str, optional): Will be used for devId
             uid(str, optional): Will be used for uid
         """
+        # dicts will get referenced instead of copied if we don't do this
+        def _deepcopy(dict1):
+            result = {}
+            for k in dict1:
+                if isinstance( dict1[k], dict ):
+                    result[k] = _deepcopy( dict1[k] )
+                else:
+                    result[k] = dict1[k]
+            return result
+
+        # dict2 will be merged into dict1
+        # as dict2 is payload_dict['...'] we only need to worry about copying 2 levels deep,
+        #  the command id and "command"/"command_override" keys: i.e. dict2[CMD_ID]["command"]
+        def _merge_payload_dicts(dict1, dict2):
+            for cmd in dict2:
+                if cmd not in dict1:
+                    # make a deep copy so we don't get a reference
+                    dict1[cmd] = _deepcopy( dict2[cmd] )
+                else:
+                    for var in dict2[cmd]:
+                        if not isinstance( dict2[cmd][var], dict ):
+                            # not a dict, safe to copy
+                            dict1[cmd][var] = dict2[cmd][var]
+                        else:
+                            # make a deep copy so we don't get a reference
+                            dict1[cmd][var] = _deepcopy( dict2[cmd][var] )
+
+        # start merging down to the final payload dict
+        # later merges overwrite earlier merges
+        # "default" - ("gateway" if gateway) - ("zigbee" if sub-device) - [version string] - ('gateway_'+[version string] if gateway) -
+        #   'zigbee_'+[version string] if sub-device - [dev_type if not "default"]
+        if not self.payload_dict or self.last_dev_type != self.dev_type:
+            self.payload_dict = {}
+            _merge_payload_dicts( self.payload_dict, payload_dict['default'] )
+            if self.children:
+                _merge_payload_dicts( self.payload_dict, payload_dict['gateway'] )
+            if self.cid:
+                _merge_payload_dicts( self.payload_dict, payload_dict['zigbee'] )
+            if self.version_str in payload_dict:
+                _merge_payload_dicts( self.payload_dict, payload_dict[self.version_str] )
+            if self.children and ('gateway_'+self.version_str) in payload_dict:
+                _merge_payload_dicts( self.payload_dict, payload_dict['gateway_'+self.version_str] )
+            if self.cid and ('zigbee_'+self.version_str) in payload_dict:
+                _merge_payload_dicts( self.payload_dict, payload_dict['zigbee_'+self.version_str] )
+            if self.dev_type != 'default':
+                _merge_payload_dicts( self.payload_dict, payload_dict[self.dev_type] )
+            log.debug( 'final payload_dict for %r (%r/%r): %r', self.id, self.version_str, self.dev_type, self.payload_dict )
+            # save it so we don't have to calculate this again unless something changes
+            self.last_dev_type = self.dev_type
+
         json_data = command_override = None
 
-        if command in payload_dict[self.dev_type]:
-            if 'command' in payload_dict[self.dev_type][command]:
-                json_data = payload_dict[self.dev_type][command]['command']
-            if 'command_override' in payload_dict[self.dev_type][command]:
-                command_override = payload_dict[self.dev_type][command]['command_override']
-
-        if self.dev_type != 'default':
-            if json_data is None and command in payload_dict['default'] and 'command' in payload_dict['default'][command]:
-                json_data = payload_dict['default'][command]['command']
-            if command_override is None and command in payload_dict['default'] and 'command_override' in payload_dict['default'][command]:
-                command_override = payload_dict['default'][command]['command_override']
+        if command in self.payload_dict:
+            if 'command' in self.payload_dict[command]:
+                json_data = self.payload_dict[command]['command']
+            if 'command_override' in self.payload_dict[command]:
+                command_override = self.payload_dict[command]['command_override']
 
         if command_override is None:
             command_override = command
@@ -1509,7 +1703,7 @@ class XenonDevice(object):
         # make sure we don't modify payload_dict
         json_data = json_data.copy()
 
-        if "gwId" in json_data or self.parent:
+        if "gwId" in json_data:
             if gwId is not None:
                 json_data["gwId"] = gwId
             elif self.parent:
@@ -1528,21 +1722,29 @@ class XenonDevice(object):
                 json_data["uid"] = self.id
         if self.cid:
             json_data["cid"] = self.cid
+            if "data" in json_data:
+                json_data["data"]["cid"] = self.cid
+                json_data["data"]["ctype"] = 0
+        #elif "cid" in json_data:
+        #    del json_data['cid']
         if "t" in json_data:
             if json_data['t'] == "int":
                 json_data["t"] = int(time.time())
             else:
                 json_data["t"] = str(int(time.time()))
-
-        if data is not None:
+        if rawData is not None and "data" in json_data:
+            json_data["data"] = rawData
+        elif data is not None:
             if "dpId" in json_data:
                 json_data["dpId"] = data
             elif "data" in json_data:
-                json_data["data"] = {"dps": data}
+                json_data["data"]["dps"] = data
             else:
                 json_data["dps"] = data
         elif self.dev_type == "device22" and command == DP_QUERY:
             json_data["dps"] = self.dps_to_request
+        if reqType and "reqType" in json_data:
+            json_data["reqType"] = reqType
 
         # Create byte buffer from hex data
         if json_data == "":
@@ -1592,9 +1794,11 @@ class Device(XenonDevice):
         log.debug("product received data=%r", data)
         return data
 
-    def heartbeat(self, nowait=False):
+    def heartbeat(self, nowait=True):
         """
-        Send a simple HEART_BEAT command to device.
+        Send a keep-alive HEART_BEAT command to keep the TCP connection open.
+
+        Devices only send an empty-payload response, so no need to wait for it.
 
         Args:
             nowait(bool): True to send without waiting for response.
@@ -1697,17 +1901,14 @@ class Device(XenonDevice):
 def pad(s):
     return s + (16 - len(s) % 16) * chr(16 - len(s) % 16)
 
-
 def unpad(s):
     return s[: -ord(s[len(s) - 1 :])]
 
-
 def encrypt(msg, key):
-    return AES.new(key, AES.MODE_ECB).encrypt(pad(msg).encode())
-
+    return AESCipher( key ).encrypt( msg, use_base64=False, pad=True )
 
 def decrypt(msg, key):
-    return unpad(AES.new(key, AES.MODE_ECB).decrypt(msg)).decode()
+    return AESCipher( key ).decrypt( msg, use_base64=False, decode_text=True )
 
 #def decrypt_gcm(msg, key):
 #    nonce = msg[:12]
@@ -1728,13 +1929,6 @@ def decrypt_udp(msg):
         return payload
     return decrypt(msg, udpkey)
 
-# Return positive number or zero
-def floor(x):
-    if x > 0:
-        return x
-    else:
-        return 0
-
 
 def appenddevice(newdevice, devices):
     if newdevice["ip"] in devices:
