@@ -68,7 +68,7 @@ SCANTIME = tinytuya.SCANTIME        # How many seconds to wait before stopping
 max_parallel = 300
 connect_timeout = 3
 
-devinfo_keys = ('ip', 'mac', 'name', 'key', 'gwId', 'active', 'ablilty', 'encrypt', 'productKey', 'version', 'token', 'wf_cfg' )
+devinfo_keys = ('ip', 'mac', 'name', 'key', 'gwId', 'active', 'ability', 'encrypt', 'productKey', 'version', 'token', 'wf_cfg' )
 # id ver
 
 TermColors = namedtuple("TermColors", "bold, subbold, normal, dim, alert, alertdim, cyan, red, yellow")
@@ -170,6 +170,10 @@ class DeviceDetect(object):
 
         if not deviceinfo:
             deviceinfo = {}
+        # some devices report "ability" but most have this as the typo "ablilty"
+        if 'ablilty' in deviceinfo and 'ability' not in deviceinfo:
+            deviceinfo['ability'] = deviceinfo['ablilty']
+            del deviceinfo['ablilty']
         self.deviceinfo = deviceinfo
         for k in devinfo_keys:
             if k not in deviceinfo:
@@ -952,7 +956,7 @@ def _print_device_info( result, note, term, extra_message=None ):
 
 
 # Scan function
-def devices(verbose=False, scantime=None, color=True, poll=True, forcescan=False, byID=False, show_timer=None, discover=True, wantips=None, wantids=None, snapshot=None, assume_yes=False, tuyadevices=[]): # pylint: disable=W0621, W0102
+def devices(verbose=False, scantime=None, color=True, poll=True, forcescan=False, byID=False, show_timer=None, discover=True, wantips=None, wantids=None, snapshot=None, assume_yes=False, tuyadevices=[], maxdevices=0): # pylint: disable=W0621, W0102
     """Scans your network for Tuya devices and returns dictionary of devices discovered
         devices = tinytuya.deviceScan(verbose)
 
@@ -967,9 +971,10 @@ def devices(verbose=False, scantime=None, color=True, poll=True, forcescan=False
         discover = True or False, when False, UDP broadcast packets will be ignored
         wantips = A list of IP addresses we want.  Scan will stop early if all are found
         wantids = A list of Device IDs we want.  Scan will stop early if all are found
-        snapshot = True or False, save snapshot once finished
+        snapshot = A dict of devices with IP addresses as keys.  These devices will be force-scanned
         assume_yes = True or False, do not prompt to confirm auto-detected network ranges
         tuyadevices = contents of devices.json, to prevent re-loading it if we already have it
+        maxdevices = Stop scanning after this many devices are found.  0 for no limit
 
     Response:
         devices = Dictionary of all devices found
@@ -1020,15 +1025,33 @@ def devices(verbose=False, scantime=None, color=True, poll=True, forcescan=False
         # Enable UDP listening broadcasting mode on UDP port 6666 - 3.1 Devices
         client = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
         client.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
+        try:
+            client.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
+        except AttributeError:
+            # SO_REUSEPORT not available
+            pass
         client.bind(("", UDPPORT))
         #client.settimeout(TIMEOUT)
+
         # Enable UDP listening broadcasting mode on encrypted UDP port 6667 - 3.3 Devices
         clients = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
         clients.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
+        try:
+            clients.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
+        except AttributeError:
+            # SO_REUSEPORT not available
+            pass
         clients.bind(("", UDPPORTS))
         #clients.settimeout(TIMEOUT)
+
+        # Enable UDP listening broadcasting mode on encrypted UDP port 7000 - App
         clientapp = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
         clientapp.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
+        try:
+            clientapp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
+        except AttributeError:
+            # SO_REUSEPORT not available
+            pass
         clientapp.bind(("", UDPPORTAPP))
     else:
         client = clients = clientapp = None
@@ -1397,6 +1420,18 @@ def devices(verbose=False, scantime=None, color=True, poll=True, forcescan=False
                     wantips.remove(ip)
                 if broadcasted_devices[ip].deviceinfo['gwId'] in wantids:
                     wantids.remove( broadcasted_devices[ip].deviceinfo['gwId'] )
+                if maxdevices:
+                    maxdevices -= 1
+                    if maxdevices == 0:
+                        if verbose:
+                            print('Found all the devices we wanted, ending scan early')
+                        ip_wantips = False
+                        ip_wantids = False
+                        ip_force_wants_end = True
+                        scan_end_time = 0
+                        for dev in devicelist:
+                            if (not dev.remove) and (not dev.passive) and ((dev.timeo + 1.0) > device_end_time):
+                                device_end_time = dev.timeo + 1.0
 
                 for dev in devicelist:
                     if dev.ip == ip:
